#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "screenhack.h"
#include "version.h"
#include "vroot.h"
#ifndef isupper
# define isupper(c)  ((c) >= 'A' && (c) <= 'Z')
#endif
#ifndef _tolower
# define _tolower(c)  ((c) - 'A' + 'a')
#endif

char *progname;
XrmDatabase db;
XtAppContext app;
Bool mono_p;

static XrmOptionDescRec default_options [] = {
  { "-root",    ".root",                XrmoptionNoArg, "True" },
  { "-window",  ".root",                XrmoptionNoArg, "False" },
  { 0, 0, 0, 0 }
};

static char *default_defaults[] = {
  ".root:		false",
  "*geometry:		600x480",
  "*installColormap:	false",
  "*visualID:		default",
  "*desktopGrabber:	xscreensaver-getimage %s",
  0
};

static XrmOptionDescRec *merged_options;
static int merged_options_size;
static char **merged_defaults;

static void
merge_options (void)
{
  int def_opts_size, opts_size;
  int def_defaults_size, defaults_size;

  for(def_opts_size=0; default_options[def_opts_size].option;
	  def_opts_size++);
  for(opts_size=0; options[opts_size].option; opts_size++);

  merged_options_size=def_opts_size+opts_size;
  merged_options=(XrmOptionDescRec*)
    malloc((merged_options_size+1)*sizeof(*default_options));
  memcpy (merged_options,default_options,
		  (def_opts_size*sizeof(*default_options)));
  memcpy (merged_options+def_opts_size, options,
		  ((opts_size+1)*sizeof(*default_options)));

  for (def_defaults_size=0; default_defaults[def_defaults_size];
       def_defaults_size++);
  for (defaults_size=0; defaults[defaults_size]; defaults_size++);

  merged_defaults=(char **)
    malloc ((def_defaults_size+defaults_size+1)*sizeof(*defaults));;
  memcpy (merged_defaults,default_defaults,
		  def_defaults_size*sizeof(*defaults));
  memcpy (merged_defaults+def_defaults_size,defaults,
		  (defaults_size+1)*sizeof(*defaults));

  {
    char **s;
    for (s=merged_defaults; *s; s++)
      if (**s=='.') {
		const char *oldr=*s;
		char *newr=(char *)malloc(strlen(oldr)+strlen(progclass)+3);
		strcpy(newr,progclass);
		strcat(newr,oldr);
		*s=newr;
	  }
  }
}

static int screenhack_ehandler(Display *dpy,XErrorEvent *error) {
  fprintf (stderr, "\nX error in %s:\n", progname);
  if (XmuPrintDefaultErrorMessage (dpy, error, stderr))
    exit (-1);
  else
    fprintf (stderr, " (nonfatal.)\n");
  return 0;
}

static Bool MapNotify_event_p(Display *dpy,XEvent *event,XPointer window) {
  return (event->xany.type==MapNotify &&
	  event->xvisibility.window==(Window) window);
}

static Atom XA_WM_PROTOCOLS, XA_WM_DELETE_WINDOW;

void screenhack_handle_event(Display *dpy,XEvent *event) {
  switch (event->xany.type) {
  case KeyPress: {
	KeySym keysym;
	char c=0;
	XLookupString (&event->xkey,&c,1,&keysym,0);
	if (c=='q'||c=='Q'||c == 3||c == 27) exit(0);
	else if (!(keysym>=XK_Shift_L&&keysym<=XK_Hyper_R)) XBell (dpy, 0);
	}
    break;
  case ButtonPress:
	XBell (dpy, 0);
	break;
  case ClientMessage:
	if (event->xclient.message_type!=XA_WM_PROTOCOLS) {
	  char *s=XGetAtomName(dpy, event->xclient.message_type);
	  if (!s) s="(null)";
	  fprintf(stderr,"%s: unknown ClientMessage %s received!\n",
			  progname,s);
	} else if(event->xclient.data.l[0]!=XA_WM_DELETE_WINDOW) {
	  char *s1=XGetAtomName(dpy,event->xclient.message_type);
	  char *s2=XGetAtomName(dpy,event->xclient.data.l[0]);
	  if (!s1) s1="(null)";
	  if (!s2) s2="(null)";
	  fprintf (stderr,"%s: unknown ClientMessage %s[%s] received!\n",
			   progname,s1,s2);
	} else exit (0);
	break;
  }
}

void screenhack_handle_events(Display *dpy) {
  while (XPending (dpy)) {
	XEvent event;
	XNextEvent (dpy, &event);
	screenhack_handle_event (dpy, &event);
  }
}


static Visual *pick_visual (Screen *screen) {
  return get_visual_resource(screen,"visualID","VisualID", False);
}

static void visual_warning(Screen *screen,Window window,
						   Visual *visual,Colormap cmap,Bool window_p) {
  char *visual_string=get_string_resource ("visualID", "VisualID");
  Visual *desired_visual=pick_visual (screen);
  char win[100];
  char why[100];

  if (window==RootWindowOfScreen (screen))
    strcpy(win,"root window");
  else
    sprintf(win,"window 0x%x",(unsigned long)window);

  if (window_p)
    sprintf (why,"-window-id 0x%x",(unsigned long)window);
  else
    strcpy (why,"-root");

  if (visual_string&&*visual_string) {
	char *s;
	for (s=visual_string; *s; s++)
	  if (isupper(*s)) *s=_tolower(*s);

	if (!strcmp (visual_string,"default") ||
		!strcmp (visual_string,"default") ||
		!strcmp (visual_string,"best"));
	else if (visual!=desired_visual) {
	  fprintf(stderr,"%s: ignoring `-visual %s' because of `%s'.\n",
			  progname,visual_string,why);
	  fprintf(stderr,"%s: using %s's visual 0x%x.\n",
			  progname,win,XVisualIDFromVisual (visual));
	}
	free(visual_string);
  }	

  if (visual==DefaultVisualOfScreen(screen)&&
      has_writable_cells(screen,visual)&&
      get_boolean_resource("installColormap","InstallColormap")) {
	fprintf (stderr, "%s: ignoring `-install' because of `%s'.\n",
			 progname,why);
	fprintf (stderr, "%s: using %s's colormap 0x%x.\n",
			 progname,win,(unsigned long)cmap);
  }
}


int main (int argc, char **argv) {
  Widget toplevel;
  Display *dpy;
  Window window;
  Screen *screen;
  Visual *visual;
  Colormap cmap;
  Bool root_p;
  XEvent event;
  Boolean dont_clear;
  char version[255];

  merge_options ();
  toplevel = XtAppInitialize (&app, progclass, merged_options,
			      merged_options_size, &argc, argv,
			      merged_defaults, 0, 0);
  dpy = XtDisplay (toplevel);
  screen = XtScreen (toplevel);
  db = XtDatabase (dpy);

  XtGetApplicationNameAndClass (dpy, &progname, &progclass);
  if (strlen (progname) >= 100) progname[100] = 0;
  XSetErrorHandler (screenhack_ehandler);
  XA_WM_PROTOCOLS = XInternAtom (dpy, "WM_PROTOCOLS", False);
  XA_WM_DELETE_WINDOW = XInternAtom (dpy, "WM_DELETE_WINDOW", False);
  {
    char *v = (char *) strdup(strchr(screensaver_id, ' '));
    char *s1, *s2, *s3, *s4;
    s1 = (char *) strchr(v,  ' '); s1++;
    s2 = (char *) strchr(s1, ' ');
    s3 = (char *) strchr(v,  '('); s3++;
    s4 = (char *) strchr(s3, ')');
    *s2 = 0;
    *s4 = 0;
    sprintf (version, "%s: from the XScreenSaver %s distribution (%s.)",
			 progclass, s1, s3);
    free(v);
  }

  if (argc > 1)
    {
      const char *s;
      int i;
      int x = 18;
      int end = 78;
      Bool help_p = !strcmp(argv[1], "-help");
      fprintf (stderr, "%s\n", version);
      for (s = progclass; *s; s++) fprintf(stderr, " ");
      fprintf (stderr, "  programed by Shusuke Kudo <kudo@soft.ics.keio.ac.jp>\n\n");

      if (!help_p)
	fprintf(stderr, "Unrecognised option: %s\n", argv[1]);
      fprintf (stderr, "Options include: ");
      for (i = 0; i < merged_options_size; i++)
	{
	  char *sw = merged_options [i].option;
	  Bool argp = (merged_options [i].argKind == XrmoptionSepArg);
	  int size = strlen (sw) + (argp ? 6 : 0) + 2;
	  if (x + size >= end)
	    {
	      fprintf (stderr, "\n\t\t ");
	      x = 18;
	    }
	  x += size;
	  fprintf (stderr, "%s", sw);
	  if (argp) fprintf (stderr, " <arg>");
	  if (i != merged_options_size - 1) fprintf (stderr, ", ");
	}
      fprintf (stderr, ".\n");
      exit (help_p ? 0 : 1);
    }

  dont_clear = get_boolean_resource ("dontClearRoot", "Boolean");
  mono_p = get_boolean_resource ("mono", "Boolean");
  if (CellsOfScreen (DefaultScreenOfDisplay (dpy)) <= 2)
    mono_p = True;

  root_p = get_boolean_resource ("root", "Boolean");

  if (root_p)
    {
      XWindowAttributes xgwa;
      window = RootWindowOfScreen (XtScreen (toplevel));
      XtDestroyWidget (toplevel);
      XGetWindowAttributes (dpy, window, &xgwa);
      cmap = xgwa.colormap;
      visual = xgwa.visual;
      visual_warning (screen, window, visual, cmap, False);
    }
  else
    {
      Boolean def_visual_p;
      visual = pick_visual (screen);
      if (toplevel->core.width <= 0)	toplevel->core.width = 600;
      if (toplevel->core.height <= 0)	toplevel->core.height = 480;

      def_visual_p = (visual == DefaultVisualOfScreen (screen));

      if (!def_visual_p)
	{
	  unsigned int bg, bd;
	  Widget new;

	  cmap = XCreateColormap (dpy, RootWindowOfScreen(screen),
				  visual, AllocNone);
	  bg = get_pixel_resource ("background", "Background", dpy, cmap);
	  bd = get_pixel_resource ("borderColor", "Foreground", dpy, cmap);

	  new = XtVaAppCreateShell (progname, progclass,
				    topLevelShellWidgetClass, dpy,
				    XtNmappedWhenManaged, False,
				    XtNvisual, visual,
				    XtNdepth, visual_depth (screen, visual),
				    XtNwidth, toplevel->core.width,
				    XtNheight, toplevel->core.height,
				    XtNcolormap, cmap,
				    XtNbackground, (Pixel) bg,
				    XtNborderColor, (Pixel) bd,
				    XtNinput, True,  /* for WM_HINTS */
				    0);
	  XtDestroyWidget (toplevel);
	  toplevel = new;
	  XtRealizeWidget (toplevel);
	  window = XtWindow (toplevel);
	}
      else
	{
	  XtVaSetValues (toplevel,
                         XtNmappedWhenManaged, False,
                         XtNinput, True,  /* for WM_HINTS */
                         0);
	  XtRealizeWidget (toplevel);
	  window = XtWindow (toplevel);

	  if (get_boolean_resource ("installColormap", "InstallColormap"))
	    {
	      cmap = XCreateColormap (dpy, window,
				   DefaultVisualOfScreen (XtScreen (toplevel)),
				      AllocNone);
	      XSetWindowColormap (dpy, window, cmap);
	    }
	  else
	    {
	      cmap = DefaultColormap (dpy, DefaultScreen (dpy));
	    }
	}
	{
	  XtPopup (toplevel, XtGrabNone);
	}
      XtVaSetValues(toplevel, XtNtitle, version, 0);
      {
        XWindowAttributes xgwa;
        XGetWindowAttributes (dpy, window, &xgwa);
        XSelectInput (dpy, window,
                      xgwa.your_event_mask | KeyPressMask | ButtonPressMask);
        XChangeProperty (dpy, window, XA_WM_PROTOCOLS, XA_ATOM, 32,
                         PropModeReplace,
                         (unsigned char *) &XA_WM_DELETE_WINDOW, 1);
      }
    }

  if (!dont_clear) {
	XSetWindowBackground(dpy,window,
	  get_pixel_resource ("background", "Background",dpy,cmap));
	XClearWindow (dpy, window);
  }

  if (!root_p)
    XIfEvent (dpy, &event, MapNotify_event_p, (XPointer) window);
  
  XSync (dpy, False);
  screenhack (dpy, window);
  return 0;
}
